#!/usr/bin/env perl -w
#
# t1d2gsx.pl
# by pts@fazekas.hu at Wed Apr  4 20:14:41 CEST 2001
#
# The .t1d file format is a human-readable text file which describes a
# (simple) complete PostScript Type1 font program. This utility (t1d2gsx.pl)
# converts the .t1d file to a .gsx font file. The .gsx font file is useless
# by itself (there doesn't exist any computer program which can load a .gsx
# file), but using the `type1fix.pl' utility from the same author one can
# convert the .gsx file to a portable .pfb file which can be used in a
# variety of printers, environments, and can be opened in font editors.
#
# This utility reads a .t1d file from STDIN, writes the .gsx file to STDOUT
# and writes debug messages to STDERR. It doesn't accept command line
# arguments (options).
#
# The .t1d file format was invented by the author of this utility in a
# spring afternoon. For professional DTP tasks, probably a better file format
# should be designed.
#
# The .t1d file format is easy to read and write for humans. The .t1d file
# is easy to generate from programs.
#
# The .t1d file format isn't documented anywhere (yet). You should read
# Adobe's t1_spec.pdf (especially the `CharStrings Dictionary' part) to get
# started.
#

#
# Imp: justlib
# Imp: $0
# Imp: quoting strings etc.
#

use integer;
use strict;

my $template=<<'END_TEMPLATE';
13 dict dup begin
  /FontInfo 16 dict dup begin
    /version ($VVV.$WWW) readonly def
    /Notice ($Notice) readonly def
    /Copyright ($Copyright) readonly def
    /FullName ($FullName) readonly def
    /FamilyName ($FamilyName) readonly def
    /Weight ($Weight) readonly def
    /ItalicAngle $ItalicAngle def
    /isFixedPitch $isFixedPitch def
    /UnderlinePosition $UnderlinePosition def
    /UnderlineThickness $UnderlineThickness def
    % /ascent 750 def
    % /descent 250 def
    % /em 1000 def
  end readonly def
  /FontName /$FontName def
  /UniqueID $UniqueID def
  /FontType 1 def
  /PaintType 0 def % 0==fill, 2==outline (stroke)
  % /StrokeWidth 0 def
  % /Metrics dictionary
  /FontMatrix [$xrevscale 0 0 $yrevscale 0 0] readonly def
  /FontBBox {$llx $lly $urx $ury} readonly def
  /Encoding 256 array
    % dup 251 /germandbls put
    $Encoding
  def
  /Private 20 dict
    % dup/ND {noaccess def} executeonly put
    % dup/NP {noaccess put} executeonly put
    % dup/RD {string currentfile exch readstring pop} executeonly put
    % dup/-! {string currentfile exch readhexstring pop} executeonly put
    % dup/| {noaccess put} executeonly put
    % dup/|- {noaccess def} executeonly put
    dup/lenIV 4 put % compatibility
    dup/password 5839 put % compatibility
    dup/MinFeature {16 16} noaccess put % compatibility
    dup/UniqueID $UniqueID put
    dup/BlueValues [] noaccess put
    % not_required: dup/OtherBlues [] noaccess put
    % not_required: dup/FamilyBlues [] noaccess put
    % not_required: dup/FamilyOtherBlues [] noaccess put
    % not_required: dup/BlueScale
    % not_required: dup/BlueShift
    % not_required: dup/BlueFuzz
    % not_required: dup/StdHW [28] noaccess put
    % not_required: dup/StdVW [84] noaccess put
    % not_required: dup/StemSnapH
    % not_required: dup/StemSnapV
    % not_required: dup/ForceBold false put
    % not_required: dup/Erode {10.5 dup 3 -1 roll 0.1 mul exch 0.5 sub mul cvi sub dup mul 84 0 dtransform dup mul exch dup mul add le {pop pop 1.0 1.0} {pop pop 0.0 1.5} ifelse} put
    % not_required: dup/ExpansionFactor 0.06 put
    % not_required: dup/RndStemUp
    dup/OtherSubrs 4 array
      dup 0 {} put
      dup 1 {} put
      dup 2 {} put
      dup 3 {3} executeonly put
    noaccess put
    dup/Subrs 1 array
      dup 0 (\020\2771p\005\266\333\321\0509) put % empty glyph
    noaccess put
  noaccess def
  /CharStrings $CharStringsLen dict dup begin
    % /.notdef {0 250 hsbw endchar} def
    % /colon (\020\2771p.A\045\310\023\273\225\076Y2\345\310\215\201i\275\273\3039\0245\222dSbL\340\3354\265QFR\076Q\240Lq\0240\025) def  
    % /colon {81 278 hsbw 25 callsubr 348 111 hstem 24 1 3 callothersubr popp callsubr 55 348 rmoveto 148 callsubr 25 1 3 callothersubr popp callsubr -359 vmoveto 148 callsubr endchar} def
    % /five {32 500 hsbw 324 70 vstem -14 37 hstem 583 79 hstem 149 583
    %   rmoveto 196 hlineto 11 0 9 0 3 9 rrcurveto 38 89 rlineto -9 7 rlineto
    %   -11 -15 -9 -11 -26 0 rrcurveto -209 hlineto -109 -237 rlineto -3 -7 3
    %   -6 6 0 rrcurveto 108 -4 177 -47 0 -160 rrcurveto -122 -74 -56 -65
    %   vhcurveto -54 0 -27 62 -64 0 rrcurveto -40 0 -29 -10 hvcurveto -36 47
    %   -24 79 vhcurveto 158 110 106 133 hvcurveto 0 89 -11 29 -43 52 rrcurveto
    %   -59 72 -94 17 -80 14 rrcurveto closepath endchar} def
    % ... but no callsubrs, please
    $CharStrings
  end readonly def
end
dup/FontName get exch definefont pop
END_TEMPLATE

my %VARS=();

sub process_gsx($) {
  # e.g process_gsx(\*STDIN)
  # updates %VARS
  my $FH=$_[0];
  while (<$FH>) {
    s/\A\s+//;
    next if /^#/ or not length;
    s/\s+\Z(?!\n)//;
    die "$0: syntax error in line $.\n" unless /(\w+)\s*(=|\+=)\s*(.*)\Z(?!\n)/;
    if ($2 eq '+=') {
      die "$0: $1 undefined in line $.\n" unless exists $VARS{$1};
      $VARS{$1}.="$3\n";
    } else {
      $VARS{$1}=$3;
    }
    # print STDERR "$1 $2 $3\n";
  }
  $VARS{CharStringsLen}=$VARS{CharStrings}
}

sub substitute($) {
  my $S=$_[0];
  $S=~s@\$(\w+)@
    die "$0: var $1 was not defined\n" unless exists $VARS{$1};
    $VARS{$1}
  @ge;
  $S
}

process_gsx(\*STDIN);
# vvv Imp: check for exists CharStrings
$VARS{CharStringsLen}=$VARS{CharStrings}=~y@/@@;
if (exists $VARS{CharBBox}) {
  # Thu Apr  5 20:41:54 CEST 2001
  my($LLX,$LLY,$URX,$URY)=(0,0,0,0);
  my $mini=sub { $_[0]=$_[1] unless $_[0]<$_[1] };
  my $maxi=sub { $_[0]=$_[1] unless $_[0]>$_[1] };
  my $C=0;
  for my $LINE (split/\n+/,$VARS{CharBBox}) {
    if ($LINE=~/^\s*(\d+)\s*(-?\d+)\s*(-?\d+)\s*(-?\d+)\s*(-?\d+)$/) {
      $mini->($LLX,$2); $mini->($LLY,$3);
      $maxi->($URX,$4); $maxi->($URY,$5);
      $C++;
    }
  }
  print STDERR "$C CharBBox entries.\n";
  die "$0: CharStringsLen must be 256, but is $VARS{CharStringsLen}\n" if @ARGV and $ARGV[0] eq '--256' and $VARS{CharStringsLen}!=256;
  if ($URX==0 and $URX==0) {
    # big enough??
    $VARS{llx}=-1000; $VARS{lly}=-1000; $VARS{urx}=2000; $VARS{ury}=2000;
  } else { 
    $VARS{llx}=$LLX; $VARS{lly}=$LLY;
    $VARS{urx}=$URX; $VARS{ury}=$URY;
  }
}
print substitute($template);

__END__

Fonts
"""""
% tir_____.pfb
name	$FontName	Times-Roman
str	$VVV		001
str	$WWW		007
str	$Notice		Copyright (c) 1985, 1987, 1989, 1990 Adobe Systems Incorporated. ....
str	$Copyright	 The digitally encoded machine readable outline data for producing...
str	$FullName	Times Roman
str	$FamilyName	Times
str	$Weight		Roman
int	$ItalicAngle	0
bool	$isFixedPitch	false
int	$UnderlinePosition	-100
int	$UnderlineThickness	50
int	$UniqueID	28427
real	$xrevscale	0.001
real	$yrevscale	0.001
int	$llx		-168
int	$lly		-218
int	$urx		1000
int	$ury		898
misc	$Encoding	...
int	$CharStringsLen	229 auto
misc	$CharStrings	...

name	$FontName	NimbusRomNo9L-ReguItal
str	$VVV		001
str	$WWW		005
str	$Notice		URW Software, Copyright 1994 by URW
str	$Copyright	Copyright URW Software, Copyright 1994 by URW
str	$FullName	Nimbus Roman No9 L Regular Italic
str	$FamilyName	Nimbus Roman No9 L
str	$Weight		Regular
int	$ItalicAngle	-15.5
bool	$isFixedPitch	false
int	$UnderlinePosition	-100
int	$UnderlineThickness	50
int	$UniqueID	5020937
real	$xrevscale	0.001
real	$yrevscale	0.001
int	$llx		-169
int	$lly		-217
int	$urx		1010
int	$ury		904
misc	$Encoding	...
int	$CharStringsLen ... auto
misc	$CharStrings	...


name	$FontName	!TeX-ecrm1000
str	$VVV		001
str	$WWW		001
str	$Notice		converted in April 2001
str	$Copyright	see copyright of original TeX font
str	$FullName	!TeX ecrm1000 Regular
str	$FamilyName	!TeX ecrm1000
str	$Weight		Regular
int	$ItalicAngle	0
bool	$isFixedPitch	false
int	$UnderlinePosition	-100
int	$UnderlineThickness	50
int	$UniqueID	!4314???
real	$xrevscale	0.001
real	$yrevscale	0.001
int	$llx		!??
int	$lly		!??
int	$urx		!??
int	$ury		!??
misc	$Encoding	!??
int	$CharStringsLen !?? auto
misc	$CharStrings	!??

Substitutions will be only near `!'.
